//
//  integrationAbscissas.cpp
//  qpso


#include "integrationAbscissas.hpp"

// Constructor Function
integrationAbscissas::integrationAbscissas():
tau(3)
{};

/*
 
 -------------------------------------------------------------------------------------------
                        performPriorConversion
 -------------------------------------------------------------------------------------------
 
 Arguments:
    priors  - vector of string objects, each element is a prior on a factor
 
 Description:
    Method to convert the string input of the priors to a usable form for the integration
    approximation.
 
 */

void integrationAbscissas::performPriorConversion(std::vector<string> priors){
    
    int numBeta = (int)priors.size();
    priorDistn.resize(numBeta);
    std::vector<string> muTemp(numBeta), varTemp(numBeta), aTemp(numBeta), bTemp(numBeta);
    
    // Resize sigma to the propert size
    sigma.resize(numBeta, numBeta);
    mu.resize(numBeta);
    a.resize(numBeta);
    b.resize(numBeta);
    for (int iRow = 0; iRow<numBeta; iRow++){
        for (int iCol = 0; iCol < numBeta; iCol++){
            sigma(iRow, iCol) = 0;
        }
    }
    
    for (int i = 0; i < numBeta; i++){
        std::size_t found = priors[i].find("(");
        std::string distn = priors[i].substr (0,found);
        
        priorDistn[i] = distn;
        
        // Handle Normal Distribution
        if ( distn  == "N" ){
            std::size_t sep = priors[i].find(",");
            muTemp[i] = priors[i].substr (found+1,sep);
            std::size_t end = priors[i].find(")");
            varTemp[i] = priors[i].substr (sep+1,end);
            // Store the results as doubles (instead of strings)
            mu(i) = std::stod (muTemp[i]);
            sigma(i,i) = std::stod (varTemp[i]);
            a(i) = 0; b(i) = 0;
        }
        
        // Handle the Uniform Distribution
        else if ( distn == "U" ){
            std::size_t sep = priors[i].find(",");
            aTemp[i] = priors[i].substr (found+1,sep);
            std::size_t end = priors[i].find(")");
            bTemp[i] = priors[i].substr (sep+1,end);
            // Store the results as doubles instead of strings
            a(i) = std::stod (aTemp[i]);
            b(i) = std::stod (bTemp[i]);
            mu(i) = 0.5 * ( a(i) + b(i) ) ;
            //mu(i) = 0 ;
            sigma(i,i) = 1.0/12 * ( (b(i) - a(i)) * (b(i) - a(i)) );
        }
        
    }
    
    cout << "the means are \n" ;
    cout << mu << endl;
    cout << "The variance covariance matrix is: " << endl;
    cout << sigma << endl;
    cout << "The starting points for the unif distn are: " << endl;
    cout << a;
    cout << "The ending points for the unif distn are: " << endl;
    cout << b;
    cout << "Thats all folks";
    
}

/*
 
 -------------------------------------------------------------------------------------------
                            generateRadialQuantities
 -------------------------------------------------------------------------------------------
 
 Arguments:
    p  - number of radii to use. Currently fixed to 3
 
 Description:
    Method to generate the radial information
 
 */

void integrationAbscissas::generateRadialQuantities(int p){
    
    radialWeights.resize(3);
    
    tau(0) = 0.0;
    radialWeights(0) = 8.0 / ( (p+2)*(p+4) );
    
    tau(1) = sqrt( p+4.0-sqrt(2*p + 8) );
    radialWeights(1) = ( p*(p+2) ) / ( (p+4)*pow((2 - sqrt(2*p+8)),2) );
    
    tau(2) = sqrt( p+4.0+sqrt(2*p + 8) );
    radialWeights(2) = ( p*(p+2) ) / ( (p+4.0)*pow((2 + sqrt(2*p+8)),2) );
    
}

/*
 
 -------------------------------------------------------------------------------------------
                            generateSimplexMethod
 -------------------------------------------------------------------------------------------
 
 Arguments:
    p  - number of parameters in the model.
 
 Description:
    Method to generate the simplex vertices and weights
 
 */

void integrationAbscissas::generateSimplexMethod(int p){
    
    // Start by forming the simplex points
    MatrixXd V(p, p+1);
    for (int iRow = 0; iRow < p; iRow++){
        for (int iCol = 0; iCol < p+1; iCol++){
            double i = iRow + 1;
            if (iRow < iCol){
                V(iRow, iCol) = -sqrt( (p+1) / (p*(p-i+2)*(p-i+1)) );
            } else if (iRow == iCol){
                V(iRow, iCol) = sqrt( ((p+1)*(p-i+1))/(p*(p-i+2)) );
            } else{
                V(iRow, iCol) = 0;
            }
        }
    }
    
    cout << "The V matrix is; \n" << V << endl;
    
    // Find the (p+1)(p+2)/2 midpoints of the simplex points
    MatrixXd midpoints(p, (p+1)*(p+2) / 2);
    VectorXd simplexWeightsTemp( (p+1)*(p+2) );
    int iMid = -1;
    double projCoefficient = sqrt( p / (2.0*(p-1)) );
    for (int iCol1 = 0; iCol1 < p+1; iCol1++){
        for (int iCol2 = iCol1; iCol2 < p+1; iCol2++){
            iMid = iMid + 1;    // increment the counter
            // Assign the weights
            if (iCol1 == iCol2){
                simplexWeightsTemp(iMid) = p*(7.0-p) / ( 2*pow(p+1, 2)*(p+2) );
                simplexWeightsTemp(iMid + ((p+1)*(p+2)/2) ) = p*(7.0-p) / ( 2*pow(p+1, 2)*(p+2) );
                midpoints.col(iMid) = V.col(iCol1);
            } else {
                simplexWeightsTemp(iMid) = 2.0*pow(p-1,2) / ( p*pow(p+1, 2)*(p+2) );
                simplexWeightsTemp(iMid + (p+1)*(p+2.0)/2 ) = 2.0*pow(p-1,2) / ( p*pow(p+1, 2)*(p+2) );
                midpoints.col(iMid) = projCoefficient * (V.col(iCol1) + V.col(iCol2));
            }
        }
    }
    
    // Get the flipped sign midpoints
    MatrixXd flippedMidpoints = -midpoints;
    
    // Assign the simplex points
    MatrixXd simplexPointsTemp(midpoints.rows(), midpoints.cols()+flippedMidpoints.cols());
    simplexPointsTemp << midpoints, flippedMidpoints;
    
    // correct -0 as zero
    for (int j = 0; j < (p+1)*(p+2); j++){
        for (int i = 0; i < p; i++){
            if (fabs(simplexPointsTemp(i,j)) == 0){
                simplexPointsTemp(i,j) = 0;
            }
        }
    }
    
    //----- Some points are repeated, combine those points and their weights -----//
    // Find the repeated points: output is maxCount amd a vector of unique points
    std::vector<int> midList( (p+1)*(p+2) );
    for (int i = 0; i < (p+1)*(p+2); i++){ midList[i] = 0; };
    int maxCount = 0;
    
    for (int iPoint1 = 0; iPoint1 < (p+1)*(p+2); iPoint1++){
        if (midList[iPoint1] == 0){
            maxCount = maxCount + 1;
            midList[iPoint1] = maxCount;
            for (int iPoint2 = iPoint1+1; iPoint2 < (p+1)*(p+2); iPoint2++){
                // loop over each row
                for (int iP = 0; iP < p; iP++){
                    if (fabs(simplexPointsTemp(iP,iPoint1)-simplexPointsTemp(iP,iPoint2)) > 0.00001){
                        iP = p;
                    }
                    // handle exact match
                    if (iP == p-1){
                        midList[iPoint2] = maxCount;
                    }
                }
            
            }   // end second point loop
        } // end check if already assigned
    }   // end first point loop
    
    // Combine the repeated points and their weights
    MatrixXd finalSimplexPoints(p,maxCount);
    VectorXd weightList = VectorXd::Zero(maxCount);

    for (int i = 0; i < maxCount; i++){ weightList(i) = 0; };
    
    for (int i = 0; i < (p+1)*(p+2); i++){
        int iColumnToEdit = midList[i] - 1;
        finalSimplexPoints.col(iColumnToEdit) = simplexPointsTemp.col(i);
        weightList(iColumnToEdit) = weightList(iColumnToEdit) + simplexWeightsTemp(i);
    }
    
    Ns = maxCount;
    simplexPoints.resize(finalSimplexPoints.rows(), finalSimplexPoints.cols());
    simplexPoints << finalSimplexPoints;
    simplexWeights.resize(maxCount);
    simplexWeights << weightList;
    
}


/*
 
 -------------------------------------------------------------------------------------------
                                generateOrthoRotations
 -------------------------------------------------------------------------------------------
 
 Arguments:
    p  - number of parameters in the model.
 
 Description:
    Method to generate the orthogonal rotation matrices.
 
 */

void integrationAbscissas::generateOrthoRotations(int p){

    Nq = 3;

    Q.resize(p*p, std::vector<double> (Nq));
    
    
    for (int i=0; i<Nq; i++){
        MatrixXd newOrthoRot(p,p);
        // Generate the standard normal matrix
        std::random_device rdm;
        std::mt19937 generator(rdm());
        std::uniform_real_distribution<double> distribution(0,1);
        for (int iRow = 0; iRow < p; iRow++){
            for (int iCol = 0; iCol < p; iCol++){
                newOrthoRot(iRow, iCol) = distribution(generator);
            }
        }
    
        // Get Q from the QR decomposition
        HouseholderQR<MatrixXd> qr(newOrthoRot);
        qr.compute(newOrthoRot);
        
        MatrixXd Qtemp = qr.matrixQR().template  triangularView<Lower>();
        Qtemp = qr.householderQ();
        
        // Store Q temp in the array
        int iPos = -1;
        for (int iRow = 0; iRow < p; iRow++){
            for (int iCol = 0; iCol < p; iCol++){
                iPos = iPos + 1;
                Q[iPos][i] = Qtemp(iRow, iCol);
            }
        }
    }
}

/*
 
 -------------------------------------------------------------------------------------------
                                generateThetaSearchPoints
 -------------------------------------------------------------------------------------------
 
 Arguments:
    p  - number of parameters in the model.
 
 Description:
    Method to generate the integration terms based on the prior (MU + L SQRT(TAUi) Qi Vk terms).
 
 */

void integrationAbscissas::generateThetaSearchPoints(int p){
    
    VectorXd beta(p);
    beta = mu;
    
    // resize thetaSearch and thetaSearchWeights to the appropriate size
    thetaSearch.resize(Nr);
    thetaSearchWeights.resize(Nr);
    for(int i = 0; i < Nr; i++)
    {
        thetaSearch[i].resize(Nq);
        thetaSearchWeights[i].resize(Ns);
        for(int j = 0; j < Nq; j++){
            thetaSearch[i][j].resize(Ns);
            for (int k = 0; k < Ns; k++)
                thetaSearch[i][j][k].resize(p);
        }
    }
    
    VectorXd muVec(p,1);
    muVec = beta;
    //cout << muVec;
    MatrixXd L(p,p);
    LLT<MatrixXd> lltOfA(sigma); // compute the Cholesky decomposition of A
    L = lltOfA.matrixL();
    
    // theta search Structure is [radial, Q, simplex, mu]
    // Blockm to handle the gauss-laguerre quadrature at 0
    for (int iP = 0; iP < p; iP++){
        thetaSearch[0][0][0][iP] = 0;
    }
    thetaSearchWeights[0][0] = radialWeights(0);
    
    // Loop over each radial quantity for remaining gauss-laguerre quadrature
    for (int iRad = 1; iRad < Nr; iRad++){

        // Loop over each Q matrix
        for (int iQ = 0; iQ<Nq; iQ++){
            
            // load the Q matrix
            MatrixXd qCurrent(p,p);
            int iPos = -1;
            for (int iRow = 0; iRow < p; iRow++){
                for (int iCol = 0; iCol < p; iCol++){
                    iPos = iPos + 1;
                    qCurrent(iRow,iCol) = Q[iPos][iQ];
                }
            }
            
            // Loop over each abscissas
            for (int iAbsc = 0; iAbsc < Ns; iAbsc++){
                // Calculate the new theta values
                VectorXd newMu(p);
                
                newMu = sqrt(tau(iRad)) * qCurrent * simplexPoints.col(iAbsc);
                // save the new theta valeus in thetaSearch
                for (int iP = 0; iP < p; iP++){
                    thetaSearch[iRad][iQ][iAbsc][iP] = newMu(iP);
                }
            }
            
            
        } // end of Q loop
        
        // Loop over the abscissas and find their weights
        for (int iAbsc = 0; iAbsc < Ns; iAbsc++){
            thetaSearchWeights[iRad][iAbsc] = (radialWeights(iRad) * simplexWeights(iAbsc)) / Nq;
        }
        
    } // end of radius loop
    
}

/*
 
 -------------------------------------------------------------------------------------------
                                phi
 -------------------------------------------------------------------------------------------
 
 Arguments:
    x  - a point
 
 Output:
    phi - value of the CDF at this point
 
 Description:
    Method to generate the cdf using John Cook's function.
 
 */

double phi(double x)
{
    // constants
    double a1 =  0.254829592;
    double a2 = -0.284496736;
    double a3 =  1.421413741;
    double a4 = -1.453152027;
    double a5 =  1.061405429;
    double p  =  0.3275911;
    
    // Save the sign of x
    int sign = 1;
    if (x < 0)
        sign = -1;
    x = fabs(x)/sqrt(2.0);
    
    // A&S formula 7.1.26
    double t = 1.0/(1.0 + p*x);
    double y = 1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*exp(-x*x);
    
    return 0.5*(1.0 + sign*y);
}

/*
 
 -------------------------------------------------------------------------------------------
                                performPriorApproximation
 -------------------------------------------------------------------------------------------
 
 Description:
    Method to perform the approximation
 
 */

void integrationAbscissas::performPriorApproximation(){
    
    double cumulativeValue = 0;
    
    // Loop over beta values and adjust the simplex points for the first info matrix
    for (int coef = 0; coef < a.size(); coef++){
        if ( priorDistn[coef] == "U" ){
            thetaSearch[0][0][0][coef] = (b[coef] + a[coef]) / 2;
        }
    }
    
    for (int i1 = 1; i1 < Nr; i1++){
        for (int i2 = 0; i2 < Nq; i2++){
            for (int i3 = 0; i3 < Ns; i3++){
                
                // Loop over beta values and adjust the simplex points
                for (int coef = 0; coef < a.size(); coef++){
                    if ( priorDistn[coef] == "U" ){
                        cumulativeValue = phi( thetaSearch[i1][i2][i3][coef] );
                        thetaSearch[i1][i2][i3][coef] = a[coef] + cumulativeValue * (b[coef] - a[coef]);
                    }
                }
                
            }
        }
    }
    
    
}

/*
 
 -------------------------------------------------------------------------------------------
                            Printing Methods
 -------------------------------------------------------------------------------------------
 
 Description:
    Methods to print out various aspects of the integration abscissas
 
 */

void integrationAbscissas::printSimplex(){
    std::cout << "\n The simplex points are:";
    for (int j = 0; j < simplexPoints.cols(); j++){
        std::cout << "\n column " << j+1 << " is: ";
        for (int i = 0; i < simplexPoints.rows(); i++){
            std::cout << simplexPoints(i,j) << " ";
        }
    }
    std::cout << " \n\n";
}

void integrationAbscissas::printQ(){
    for (int i =0; i<Nq; i++){
        std::cout << "The " << i+1 << "th Q matrix is: \n";
        for (int iRow = 0; iRow < 3; iRow++){
            for (int iCol = 0; iCol < 3; iCol++){
                std::cout << Q[3*iRow + iCol][i] << " ";
            }
            std::cout << "\n";
        }
        std::cout << " \n\n";
        
    }
}




